;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Soft Scheme - Copyright (c) 1993,1994 by Andrew K. Wright
;; See LICENSE for a full description of the license terms.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Lexical Binding
;;
;; The binding occurrence and all bound occurrences of a variable
;; are replaced by a single shared record.  The sharing represents
;; the binding information.
;;
;; These routines use state to avoid passing around the environment
;; under construction.

(define n-unbound 0)

(define bind-defs
  (lambda (defs env0 tenv0 old-unbound timestamp)
    (letrec
      ((cons-mutable #f)
       (unbound '())
       (use-var
        (lambda (x env context mk-node)
          (match (lookup? env x)
            (#f (let* ((b (bind-var x))
                       (n (mk-node b)))
                  (set-Name-timestamp! b context)
                  (set! unbound (cons n unbound))
                  n))
            (b (when (and (Name-primitive b) (memq x cons-mutators))
                 (set! cons-mutable #t))
               (set-Name-occ! b (+ 1 (Name-occ b)))
               (mk-node b)))))
       (bind-var
        (lambda (x)
          (make-Name x #f timestamp 0 #f #f #f #f #f #f #f #f)))
       (bind
        (lambda (e env tenv context)
          (let ((bind-cur (lambda (x) (bind x env tenv context))))
            (match e
              (($ Var x)
               (use-var x env context make-Var))
              (($ Prim x)
               (use-var x initial-env context make-Var))
              (($ Const c pred)
               (use-var pred initial-env context (lambda (p) (make-Const c p))))
              (($ Lam args e2)
               (let* ((b-args (map bind-var args))
                      (newenv (extend-env* env args b-args)))
                 (make-Lam b-args (bind e2 newenv tenv context))))
              (($ Vlam args rest e2)
               (let* ((b-args (map bind-var args))
                      (b-rest (bind-var rest))
                      (newenv (extend-env* env
                                (cons rest args) (cons b-rest b-args))))
                 (make-Vlam b-args b-rest (bind e2 newenv tenv context))))
              (($ Match e1 clauses)
               (make-Match
                 (bind-cur e1)
                 (map (lambda (x) (bind-mclause x env tenv context)) clauses)))
              (($ App e1 args)
               (make-App (bind-cur e1) (map bind-cur args)))
              (($ Begin exps)
               (make-Begin (map bind-cur exps)))
              (($ And exps)
               (make-And (map bind-cur exps)))
              (($ Or exps)
               (make-Or (map bind-cur exps)))
              (($ If test then els)
               (make-If (bind-cur test) (bind-cur then) (bind-cur els)))
              (($ Delay e2)
               (make-Delay (bind-cur e2)))
              (($ Set! x e2)
               (use-var x env context
                 (lambda (b)
                   (when (Name-struct b)
                     (syntax-err (pexpr e)
                       "define-structure identifier ~a may not be assigned" x))
                   (when (Name-primitive b)
                     (syntax-err (pexpr e)
                       "(set! ~a ...) requires (define ~a ...)" x x))
                   (when (and (not (Name-mutated b))
                              (not (= (Name-timestamp b) timestamp)))
                     (syntax-err (pexpr e)
                       "(set! ~a ...) missing from compilation unit defining ~a"
                       x x))
                   (set-Name-mutated! b #t)
                   (make-Set! b (bind-cur e2)))))
              (($ Let args e2)
               (let* ((b-args (map (match-lambda
                                     (($ Bind x e)
                                      (make-Bind (bind-var x) (bind-cur e))))
                                   args))
                      (newenv (extend-env* env
                                (map Bind-name args) (map Bind-name b-args))))
                 (make-Let b-args (bind e2 newenv tenv context))))
              (($ Let* args e2)
               (recur loop ((args args)(b-args '())(env env))
                 (match args
                   ((($ Bind x e) . rest)
                    (let ((b (bind-var x)))
                      (loop rest
                            (cons (make-Bind b (bind e env tenv context)) b-args)
                            (extend-env env x b))))
                   (() (make-Let* (reverse b-args) (bind e2 env tenv context))))))
              (($ Letr args e2)
               (let* ((b-args (map (match-lambda
                                     (($ Bind x e) (make-Bind (bind-var x) e)))
                                   args))
                      (newenv (extend-env* env
                                (map Bind-name args) (map Bind-name b-args)))
                      (b-args (map (match-lambda
                                     (($ Bind b e)
                                      (let* ((n (Name-occ b))
                                             (e2 (bind e newenv tenv context)))
                                        ; dont include self recursive refs
                                        (set-Name-occ! b n)
                                        (make-Bind b e2))))
                                   b-args)))
                 (make-Letr b-args (bind e2 newenv tenv context))))
              (($ Body defs exps)
               (match-let* (((defs newenv newtenv)
                             (bind-defn defs env tenv #f)))
                 (make-Body defs
                   (map (lambda (x) (bind x newenv newtenv context)) exps))))
              (($ Record args)
               (make-Record
                 (map (match-lambda
                        (($ Bind x e)
                         (new-field! x)
                         (make-Bind x (bind-cur e))))
                      args)))
              (($ Field x e2)
               (new-field! x)
               (make-Field x (bind-cur e2)))
              (($ Cast ty e2)
               (match-let (((t absv)
                            (handle
                              (R+collect tenv
                                (match ty
                                  (('rec bind ty2)
                                   `(rec ,bind (,ty2 -> ,ty2)))
                                  (_
                                   `(,ty -> ,ty))))
                              (match-lambda*
                                (('type . args) (apply syntax-err ty args))
                                (x (apply raise x))))))
                 (make-Cast (list ty t absv) (bind-cur e2))))))))
       (bind-mclause
        (lambda (clause env tenv context)
          (match-let*
            ((($ Mclause pattern body failsym) clause)
             (patenv empty-env)
             (bp (recur loop ((p pattern))
                   (match p
                     (($ Pvar x)
                      (when (bound? patenv x)
                        (syntax-err (ppat pattern)
                          "pattern variable ~a repeated" x))
                      (let ((b (bind-var x)))
                        (set! patenv (extend-env patenv x b))
                        (make-Pvar b)))
                     (($ Pobj c args)
                      (use-var c env context
                        (lambda (b)
                          (cond ((boolean? (Name-predicate b))
                                 (syntax-err (ppat pattern)
                                   "~a is not a predicate" c))
                                ((and (not (eq? b %vector?))
                                      (not (eq? b %cvector?))
                                      (not (= (length (cdr (Name-predicate b)))
                                              (length args))))
                                 (syntax-err (ppat pattern)
                                   "~a requires ~a sub-patterns"
                                   c (length (cdr (Name-predicate b)))))
                                (else
                                 (make-Pobj b (map loop args)))))))
                     (($ Pand pats)
                      (make-Pand (map loop pats)))
                     (($ Pnot pat)
                      (make-Pnot (loop pat)))
                     (($ Ppred pred)
                      (use-var pred env context
                        (lambda (b)
                          (unless (Name-predicate b)
                            (syntax-err (ppat pattern)
                              "~a is not a predicate" pred))
                          (make-Ppred b))))
                     (($ Pany) p)
                     (($ Pelse) p)
                     (($ Pconst c pred)
                      (use-var pred initial-env context
                        (lambda (p) (make-Pconst c p))))))))
            (if failsym
                (let ((b (bind-var failsym)))
                  (when (bound? patenv failsym)
                    (syntax-err (ppat pattern)
                      "fail symbol ~a repeated" failsym))
                  (set! patenv (extend-env patenv failsym b))
                  (make-Mclause
                    bp
                    (bind body (join-env env patenv) tenv context)
                    b))
                (make-Mclause
                  bp
                  (bind body (join-env env patenv) tenv context)
                  #f)))))
       (bind-defn
        (lambda (defs env tenv glob)
          (let*
            ((newenv empty-env)
             (newtenv empty-env)
             (struct-def (lambda (x pure)
                           (when (or (bound? newenv x)
                                     (and glob (bound? initial-env x)))
                             (syntax-err #f "~a defined more than once" x))
                           (let ((b (bind-var x)))
                             (set-Name-primitive! b #t)
                             (set-Name-struct! b #t)
                             (set-Name-pure! b pure)
                             (set! newenv (extend-env newenv x b))
                             b)))
             (bind1 (match-lambda
                      ((and z ($ Define x e))
                       (cond ((not x)
                              z)
                             ((bound? newenv x)
                              (if glob
                                  (make-Define #f (make-Set! x e))
                                  (syntax-err #f "~a defined more than once" x)))
                             (else
                              (let ((b (bind-var x)))
                                (set-Name-gdef! b glob)
                                (set! newenv (extend-env newenv x b))
                                (make-Define b e)))))
                      ((and d ($ Defstruct tag args make pred get set getn setn mutable))
                       (let* ((make (struct-def make (map not mutable)))
                              (pred (struct-def pred #t))
                              (bind-get
                               (lambda (name n)
                                 (match name
                                   (($ Some x)
                                    (let ((b (struct-def x #t)))
                                      (set-Name-selector! b
                                        (lambda (X)
                                          (make-Pobj
                                            pred
                                            (map-with-n (lambda (_ m)
                                                          (if (= m n)
                                                              X
                                                              (make-Pany)))
                                              get))))
                                      (Some b)))
                                   (none none))))
                              (bind-set
                               (match-lambda
                                 (($ Some x) (Some (struct-def x #t)))
                                 (none none)))
                              (get (map-with-n bind-get get))
                              (getn (map-with-n bind-get getn))
                              (set (map bind-set set))
                              (setn (map bind-set setn))
                              (_ (when (bound? newtenv tag)
                                   (syntax-err (pdef d)
                                     "type constructor ~a defined more than once" tag)))
                              (tc (bind-tycon
                                    tag
                                    mutable
                                    (bound? tenv tag)
                                    (lambda args
                                      (apply syntax-err (cons (pdef d) args))))))
                         (set! newtenv (extend-env newtenv tag tc))
                         (set-Name-predicate! pred
                           `(,tc ,@(map (lambda (_) (gensym)) get)))
                         (make-Defstruct tc args make pred get set getn setn mutable)))
                      ((and d ($ Datatype dt))
                       (make-Datatype
                         (mapLR
                           (match-lambda
                             (((tag . args) . bindings)
                              (when (bound? newtenv tag)
                                (syntax-err (pdef d)
                                  "type constructor ~a defined more than once" tag))
                              (let ((tc (bind-tycon
                                          tag
                                          (map (lambda (_) #f) args)
                                          (bound? tenv tag)
                                          (lambda args
                                            (apply syntax-err (cons (pdef d) args))))))
                                (set! newtenv (extend-env newtenv tag tc))
                                (cons (cons tc args)
                                      (mapLR
                                        (match-lambda
                                          (($ Variant con pred arg-types)
                                           (let ((make (struct-def con #t))
                                                 (pred (struct-def pred #t)))
                                             (set-Name-predicate! pred (cons tc args))
                                             (set-Name-variant! pred arg-types)
                                             (make-Variant make pred arg-types))))
                                        bindings)))))
                           dt)))))
             (defs2 (mapLR bind1 defs))
             (newenv2 (join-env env newenv))
             (newtenv2 (join-env tenv newtenv))
             (bind2 (match-lambda
                      ((and ($ Define (? Name? x) ($ Var y)))
                       (=> fail)
                       (if (eq? (Name-name x) y)
                           (if (bound? initial-env y)
                               (make-Define x (make-Var (lookup initial-env y)))
                               (begin
                                 (printf "Warning: (define ~a ~a) but ~a is not a primitive~%" y y y)
                                 (fail)))
                           (fail)))
                      ((and ($ Define x e2) context)
                       (when (and glob (Name? x) (bound? initial-env (Name-name x)))
                         (printf "Note: (define ~a ...) hides primitive ~a~%"
                           (Name-name x) (Name-name x)))
                       (make-Define
                         (or x (let ((b (bind-var x)))
                                 (set-Name-gdef! b glob)
                                 b))
                         (bind e2 newenv2 newtenv2 context)))
                      (d d))))
            (list (mapLR bind2 defs2) newenv2 newtenv2))))
       (bind-old
        (lambda (e env)
          (match e
            (($ Var x)
             (match (lookup? env (Name-name x))
               (#f (set! unbound (cons e unbound)))
               (b (when (and (Name-primitive b) (memq x cons-mutators))
                    (set! cons-mutable #t))
                  (set-Name-occ! b (+ 1 (Name-occ b)))
                  (set-Var-name! e b))))
            (($ Set! x _)
             (match (lookup? env (Name-name x))
               (#f (set! unbound (cons e unbound)))
               (b (when (Name-struct b)
                    (syntax-err (pexpr e)
                      "define-structure identifier ~a may not be assigned" x))
                  (when (Name-primitive b)
                    (syntax-err (pexpr e)
                      "(set! ~a ...) requires (define ~a ...)" x x))
                  (when (and (not (Name-mutated b))
                             (not (= (Name-timestamp b) timestamp)))
                    (syntax-err (pexpr e)
                      "(set! ~a ...) missing from compilation unit defining ~a"
                      x x))
                  (set-Name-mutated! b #t)
                  (set-Name-occ! b (+ 1 (Name-occ b)))
                  (set-Set!-name! e b))))))))
      (match-let (((defs env tenv)
                   (bind-defn defs env0 tenv0 #t)))
        (for-each (lambda (x) (bind-old x env)) old-unbound)
        (set-cons-mutability! cons-mutable)
        (set! n-unbound (length unbound))
        (list defs env tenv unbound)))))

(define rebind-var
  (lambda (b)
    (make-Name
      (Name-name b)
      (Name-ty b)
      (Name-timestamp b)
      (Name-occ b) ; inaccurate
      (Name-mutated b)
      #f
      #f
      #f
      #f
      #f
      #f
      #f)))

(define warn-unbound
  (lambda (l)
    (let* ((names '())
           (node->name (match-lambda
                         (($ Var x) x)
                         (($ Set! x _) x)
                         (($ Pobj x _) x)
                         (($ Ppred x) x)))
           (warn (lambda (b)
                   (unless (memq (Name-name b) names)
                     (set! names (cons (Name-name b) names))
                     (printf "Warning: ~a is unbound in " (Name-name b))
                     (print-context (pexpr (Name-timestamp b)) 2)))))
      (for-each (lambda (x) (warn (node->name x))) l))))

(define Name-unbound?
  (lambda (x)
    (not (number? (Name-timestamp x)))))
