;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Soft Scheme - Copyright (c) 1993,1994 by Andrew K. Wright
;; See LICENSE for a full description of the license terms.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Drivers

(define tree '())
(define global-env empty-env)

(define verbose #f)
(define times #f)
(define benchmarking #f)

(define cons-mutators '(set-car! set-cdr!))

(define st:check
  (lambda args
    (parameterize ((print-level #f)
                   (print-length #f)
                   (pretty-maximum-lines #f))
      (let ((output (apply do-soft args)))
        (when output
          (printf "Typed program written to file ~a~%" output))))))
    
(define st:run
  (lambda (file)
    (parameterize ((optimize-level 3))
      (when benchmarking
        (printf "Reloading slow CHECKs...~%")
        (load (string-append installation-directory "checklib.scm"))
        (set! benchmarking #f))
      (load file))))

(define st:bench
  (lambda (file)
    (parameterize ((optimize-level 3))
      (unless benchmarking
        (unless fastlibrary-file
          (use-error "No benchmarking mode in this version"))
        (printf "Reloading fast CHECKs...~%")
        (load (string-append installation-directory fastlibrary-file))
        (set! benchmarking #t))
      (load file))))

(define st:
  (lambda args
    (parameterize ((print-level #f)
                   (print-length #f)
                   (pretty-maximum-lines #f))
      (let ((output (apply do-soft args)))
        (cond
          ((not output)
           (use-error "Output file name required to run"))
          ((= 0 n-unbound)
           (printf "Typed program written to file ~a, executing ...~%" output)
           (flush-output)
           (st:run output))
          (else
           (printf "Typed program written to file ~a, not executing (unbound refs)~%" output)))))))

(define do-soft
  ; returns the output file name
  (match-lambda*
    ((input (? string? output))
     (when (strip-suffix output)
       (use-error "output file name cannot end in .ss or .scm"))
     (cond ((string? input)
            (soft-files (list input) output)
            output)
           ((and (list? input) (andmap string? input))
            (soft-files input output)
            output)
           (else
            (soft-def input output)
            output)))
    ((input #f)
     (cond ((string? input)
            (soft-files (list input) #f)
            #f)
           ((and (list? input) (andmap string? input))
            (soft-files input #f)
            #f)
           (else
            (soft-def input #f)
            #f)))
    ((input)
     (cond ((string? input)
            (let ((o (string-append (or (strip-suffix input) input) ".soft")))
              (soft-files (list input) o)
              o))
           ((and (list? input) (andmap string? input))
            (use-error "Output file name required"))
           (else
            (soft-def input #t)
            #f)))
    (else
     (use-error "Input must be a file name or list of file names"))))

(define Rawmode #f)

(define st:control
  (lambda args
    (let ((dbg (match-lambda
                 ('raw (set! display-type ptype) (set! Rawmode #t))
                 ('!raw (set! display-type tidy) (set! Rawmode #f))
                 ('verbose (set! verbose #t))
                 ('!verbose (set! verbose #f))
                 ('times (set! times #t))
                 ('!times (set! times #f))
                 ('partial (set! FullSharing #f))
                 ('!partial (set! FullSharing #t))
                 ('pseudo (set! Pseudo pseudo-subtype))
                 ('!pseudo (set! Pseudo #f))
                 ('populated (set! Populated #t))
                 ('!populated (set! Populated #f))
                 ('matchst (set! Matchst #t))
                 ('!matchst (set! Matchst #f))
                 ('genmatch (set! GenMatch #t))
                 ('!genmatch (set! GenMatch #f))
                 ('letonce (set! LetOnce #t))
                 ('!letonce (set! LetOnce #f))
                 ('global-error (set! Global-error #t))
                 ('!global-error (set! Global-error #f))
                 ('share (set! Share #t))
                 ('!share (set! Share #f))
                 ('flags (set! Flags #t))
                 ('!flags (set! Flags #f))
                 ('depths (set! Dump-depths #t))
                 ('!depths (set! Dump-depths #f))
                 ('match (set! Keep-match #t))
                 ('!match (set! Keep-match #f))
                 (x
                  (printf "Error: unknown debug switch ~a~%" x)
                  (st:control)))))
      (if (null? args)
          (begin
            (printf "Current values:")
            (for-each (lambda (x) (printf " ~a" x)) (show-controls))
            (printf "~%"))
          (for-each dbg args)))))

(define show-controls
  (lambda ()
    (list
      (if Rawmode 'raw '!raw)
      (if verbose 'verbose '!verbose)
      (if times 'times '!times)
      (if Share 'share '!share)
      (if Flags 'flags '!flags)
      (if Dump-depths 'depths '!depths)
      (if FullSharing '!partial 'partial)
      (if Pseudo 'pseudo '!pseudo)
      (if Populated 'populated '!populated)
      (if LetOnce 'letonce '!letonce)
      (if Matchst 'matchst '!matchst)
      (if GenMatch 'genmatch '!genmatch)
      (if Global-error 'global-error '!global-error)
      (if Keep-match 'match '!match))))

(define soft-def                 ; type a definition
  (lambda (exp output)
    (reinit-macros!)
    (reinit-types!)
    (reinit-output!)
    (set! visible-time 0)
    (match-let*
      ((before-parse (cpu-time))
       (defs (parse-def exp))
       (before-bind (cpu-time))
       ((defs env tenv unbound)
        (bind-defs defs initial-env initial-type-env '() 0))
       (_ (warn-unbound unbound))
       (_ (if cons-is-mutable
              (printf "Note: use of ~a, treating cons as MUTABLE~%"
                cons-mutators)
              (printf "Note: no use of ~a, treating cons as immutable~%"
                cons-mutators)))
       (before-improve (cpu-time))
       (defs (improve-defs defs))
       (before-typecheck (cpu-time))
       (_ (type-check defs))
       (_ (set! global-env env))
       (before-output (cpu-time))
       (_ (check output))
       (_ (print-summary ""))
       (before-end (cpu-time)))
      (when times
        (printf "~a seconds parsing,~%"
          (exact->inexact (* (- before-bind before-parse) clock-granularity)))
        (printf "~a seconds binding,~%"
          (exact->inexact (* (- before-improve before-bind) clock-granularity)))
        (printf "~a seconds improving,~%"
          (exact->inexact (* (- before-typecheck before-improve) clock-granularity)))
        (printf "~a seconds type checking,~%"
          (exact->inexact (* (- (- before-output before-typecheck) visible-time) clock-granularity)))
        (printf "~a seconds setting visibility,~%"
          (exact->inexact (* visible-time clock-granularity)))
        (printf "~a seconds writing output,~%"
          (exact->inexact (* (- before-end before-output) clock-granularity)))
        (printf "~a seconds in total.~%"
          (exact->inexact (* (- before-end before-parse) clock-granularity)))))))

(define type-check
  (lambda (defs)
    (set! tree defs)
    (type-defs defs)
    defs))

(define soft-files               ; type a bunch of files
  (lambda (files output)
    (let ((contents (map (lambda (f) `(begin ,@(readfile f))) files)))
      (soft-def `(begin ,@contents) output))))

(define strip-suffix
  (lambda (name)
    (let ((n (string-length name)))
      (or (and (<= 3 n) (equal? ".ss" (substring name (- n 3) n))
               (substring name 0 (- n 3)))
          (and (<= 4 n) (equal? ".scm" (substring name (- n 4) n))
               (substring name 0 (- n 4)))))))

(define st:deftype
  (match-lambda*
    (((? symbol? x) . (? list? mutability))
     (=> fail)
     (if (andmap boolean? mutability)
         (deftype x mutability)
         (fail)))
    (args (use-error "Invalid command ~a" `(st:deftype ,@args)))))

(define st:defprim
  (match-lambda*
    (((? symbol? x) type)
     (defprim x type 'impure))
    (((? symbol? x) type (? symbol? mode))
     (defprim x type mode))
    (args (use-error "Invalid command ~a" `(st:defprim ,@args)))))

(define st:help
  (lambda ()
    (printf "Commands for Soft Scheme (~a)~%" st:version)
    (printf "  (st:         file (output))    type check file and execute~%")
    (printf "  (st:type     (name))           print types of global defs~%")
    (printf "  (st:check    file (output))    type check file~%")
    (printf "  (st:run      file)             execute type checked file~%")
    (printf "  (st:bench    file)             execute type checked file fast~%")
    (printf "  (st:ltype    (name))           print types of local defs~%")
    (printf "  (st:cause)                     print cause of CHECKs~%")
    (printf "  (st:summary)                   print summary of CHECKs~%")
    (printf "  (st:help)                      prints this message~%")
    (printf "  (st:defprim  name type (mode)) define a new primitive~%")
    (printf "  (st:deftype  name bool ...)    define a new type constructor~%")
    (printf "  (st:control  flag ...)         set internal flags~%")
    (printf "For more info, see ftp://ftp.nj.nec.com/pub/wright/ssmanual/softscheme.html~%")
    (printf "Copyright (c) 1993, 1994, 1995 by Andrew K. Wright under the~%")
    (printf "terms of the Gnu Public License. No warranties of any kind apply.~%")
    ))

(define st:type type)
(define st:ltype localtype)
(define st:cause cause)
(define st:summary (lambda () (print-summary "")))

(define init!
  (lambda ()
    (when customization-file
      (load (string-append installation-directory customization-file)))
    (let ((softrc (string-append home-directory "/.softschemerc")))
      (when (file-exists? softrc)
        (load softrc)))
    (set! global-env initial-env)
    (st:help)))
