;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Soft Scheme - Copyright (c) 1993,1994 by Andrew K. Wright
;; See LICENSE for a full description of the license terms.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Initial environments, etc.

(define qop (lambda (s) (string->symbol (string-append "# " s))))
(define Qcons (qop "cons"))
(define Qbox (qop "box"))
(define Qlist (qop "list"))
(define Qvector (qop "vector"))

;; All primitives must be procedures.

(define initial-info
  ;  name           type        optional-attribute(s)
  ; where optional-attribute(s) are any of:
  ;  (S selector-pattern)   is a selector for match
  ;  (P predicate-pattern)  is a predicate for match
  ;  (I)                    impure
  ;  (Ic)                   impure if cons is impure
  ;  (D)                    no DEFINITE check
  ;  (N)                    no check at all
  ;  (C)                    implementation contains a check
  ; Note: things with mutable inputs are also marked impure.
  `(; booleans
    (not            (a -> bool)                                          )

    ; equivalence predicates
    (eqv?           (a a -> bool)                                        )
    (eq?            (a a -> bool)                                        )
    (equal?         (a a -> bool)                                        )

    ; pairs and lists
    (cons           (a b -> (cons a b))                                  (Ic))
    (car            ((cons a b) -> a)                                    (S (X . _)))
    (cdr            ((cons b a) -> a)                                    (S (_ . X)))
    (caar           ((cons (cons a b) c) -> a)                           (S ((X . _) . _)))
    (cadr           ((cons c (cons a b)) -> a)                           (S (_ X . _)))
    (cdar           ((cons (cons b a) c) -> a)                           (S ((_ . X) . _)))
    (cddr           ((cons c (cons b a)) -> a)                           (S (_ _ . X)))
    (caaar          ((cons (cons (cons a b) c) d) -> a)                  (S (((X . _) . _) . _)))
    (caadr          ((cons d (cons (cons a b) c)) -> a)                  (S (_ (X . _) . _)))
    (cadar          ((cons (cons c (cons a b)) d) -> a)                  (S ((_ X . _) . _)))
    (caddr          ((cons d (cons c (cons a b))) -> a)                  (S (_ _ X . _)))
    (cdaar          ((cons (cons (cons b a) c) d) -> a)                  (S (((_ . X) . _) . _)))
    (cdadr          ((cons d (cons (cons b a) c)) -> a)                  (S (_ (_ . X) . _)))
    (cddar          ((cons (cons c (cons b a)) d) -> a)                  (S ((_ _ . X) . _)))
    (cdddr          ((cons d (cons c (cons b a))) -> a)                  (S (_ _ _ . X)))
    (caaaar         ((cons (cons (cons (cons a b) c) d) e) -> a)         (S ((((X . _) . _) . _) . _)))
    (caaadr         ((cons e (cons (cons (cons a b) c) d)) -> a)         (S (_ ((X . _) . _) . _)))
    (caadar         ((cons (cons d (cons (cons a b) c)) e) -> a)         (S ((_ (X . _) . _) . _)))
    (caaddr         ((cons e (cons d (cons (cons a b) c))) -> a)         (S (_ _ (X . _) . _)))
    (cadaar         ((cons (cons (cons c (cons a b)) d) e) -> a)         (S (((_ X . _) . _) . _)))
    (cadadr         ((cons e (cons (cons c (cons a b)) d)) -> a)         (S (_ (_ X . _) . _)))
    (caddar         ((cons (cons d (cons c (cons a b))) e) -> a)         (S ((_ _ X . _) . _)))
    (cadddr         ((cons e (cons d (cons c (cons a b)))) -> a)         (S (_ _ _ X . _)))
    (cdaaar         ((cons (cons (cons (cons b a) c) d) e) -> a)         (S ((((_ . X) . _) . _) . _)))
    (cdaadr         ((cons e (cons (cons (cons b a) c) d)) -> a)         (S (_ ((_ . X) . _) . _)))
    (cdadar         ((cons (cons d (cons (cons b a) c)) e) -> a)         (S ((_ (_ . X) . _) . _)))
    (cdaddr         ((cons e (cons d (cons (cons b a) c))) -> a)         (S (_ _ (_ . X) . _)))
    (cddaar         ((cons (cons (cons c (cons b a)) d) e) -> a)         (S (((_ _ . X) . _) . _)))
    (cddadr         ((cons e (cons (cons c (cons b a)) d)) -> a)         (S (_ (_ _ . X) . _)))
    (cdddar         ((cons (cons d (cons c (cons b a))) e) -> a)         (S ((_ _ _ . X) . _)))
    (cddddr         ((cons e (cons d (cons c (cons b a)))) -> a)         (S (_ _ _ _ . X)))
    (set-car!       ((cons a b) a -> void)                               )
    (set-cdr!       ((cons a b) b -> void)                               )
    (list           ((&list a) -> (list a))                              (Ic))
    (length         ((list a) -> num)                                    )
    (append         ((&list (list a)) -> (list a))                       (Ic) (D))
    (reverse        ((list a) -> (list a))                               (Ic))
    (list-tail      ((list a) num -> (list a))                                (C))
    (list-ref       ((list a) num -> a)                                       (C))
    (memq           (a (list a) -> (+ false (cons a (list a))))          )
    (memv           (a (list a) -> (+ false (cons a (list a))))          )
    (member         (a (list a) -> (+ false (cons a (list a))))          )
    (assq           (a (list (cons a c)) -> (+ false (cons a c)))        )
    (assv           (a (list (cons a c)) -> (+ false (cons a c)))        )
    (assoc          (a (list (cons a c)) -> (+ false (cons a c)))        )

;    types for faster versions that do less argument checking
;    (list-tail      ((MU x (+ nil (cons a x) b)) num ->
;                        (MU y (+ nil (cons a y) b)))                    )
;    (list-ref       ((MU x (+ nil (cons a x) b)) num -> a)              )
;    (memq           (a (MU x (+ nil (cons b x) c)) ->
;                        (+ false (cons b (MU y (+ nil (cons b y) c))))) )
;    (memv           (a (MU x (+ nil (cons b x) c)) ->
;                        (+ false (cons b (MU y (+ nil (cons b y) c))))) )
;    (member         (a (MU x (+ nil (cons b x) c)) ->
;                        (+ false (cons b (MU y (+ nil (cons b y) c))))) )
;    (assq           (a (MU x (+ nil (cons (+ (cons b c) d) x) e)) ->
;                        (+ false (cons b c)))                           )
;    (assv           (a (MU x (+ nil (cons (+ (cons b c) d) x) e)) ->
;                        (+ false (cons b c)))                           )
;    (assoc          (a (MU x (+ nil (cons (+ (cons b c) d) x) e)) ->
;                        (+ false (cons b c)))                           )
    
    ; symbols
    (symbol->string (sym -> str)                                         )
    (string->symbol (str -> sym)                                         )

    ; numbers
    (complex?       (a -> bool)                                          )
    (real?          (a -> bool)                                          )
    (rational?      (a -> bool)                                          )
    (integer?       (a -> bool)                                          )
    (exact?         (num -> bool)                                        )
    (inexact?       (num -> bool)                                        )
    (=              (num num (&list num) -> bool)                        )
    (<              (num num (&list num) -> bool)                        )
    (>              (num num (&list num) -> bool)                        )
    (<=             (num num (&list num) -> bool)                        )
    (>=             (num num (&list num) -> bool)                        )
    (zero?          (num -> bool)                                        )
    (positive?      (num -> bool)                                        )
    (negative?      (num -> bool)                                        )
    (odd?           (num -> bool)                                        )
    (even?          (num -> bool)                                        )
    (max            (num (&list num) -> num)                             )
    (min            (num (&list num) -> num)                             )
    (+              ((&list num) -> num)                                 )
    (*              ((&list num) -> num)                                 )
    (-              (num (&list num) -> num)                             )
    (/              (num (&list num) -> num)                             )
    (abs            (num -> num)                                         )
    (quotient       (num num -> num)                                     )
    (remainder      (num num -> num)                                     )
    (modulo         (num num -> num)                                     )
    (gcd            ((&list num) -> num)                                 )
    (lcm            ((&list num) -> num)                                 )
    (numerator      (num -> num)                                         )
    (denominator    (num -> num)                                         )
    (floor          (num -> num)                                         )
    (ceiling        (num -> num)                                         )
    (truncate       (num -> num)                                         )
    (round          (num -> num)                                         )
    (rationalize    (num num -> num)                                     )
    (exp            (num -> num)                                         )
    (log            (num -> num)                                         )
    (sin            (num -> num)                                         )
    (cos            (num -> num)                                         )
    (tan            (num -> num)                                         )
    (asin           (num -> num)                                         )
    (acos           (num -> num)                                         )
    (atan           (num (&optional num) -> num)                         )
    (sqrt           (num -> num)                                         )
    (expt           (num num -> num)                                     )
    (make-rectangular (num num -> num)                                   )
    (make-polar     (num num -> num)                                     )
    (real-part      (num -> num)                                         )
    (imag-part      (num -> num)                                         )
    (magnitude      (num -> num)                                         )
    (angle          (num -> num)                                         )
    (exact->inexact (num -> num)                                         )
    (inexact->exact (num -> num)                                         )
    (number->string (num (&optional num) -> str)                         )
    (string->number (str (&optional num) -> num)                         )

    ; characters
    (char=?         (char char -> bool)                                  )
    (char<?         (char char -> bool)                                  )
    (char>?         (char char -> bool)                                  )
    (char<=?        (char char -> bool)                                  )
    (char>=?        (char char -> bool)                                  )
    (char-ci=?      (char char -> bool)                                  )
    (char-ci<?      (char char -> bool)                                  )
    (char-ci>?      (char char -> bool)                                  )
    (char-ci<=?     (char char -> bool)                                  )
    (char-ci>=?     (char char -> bool)                                  )
    (char-alphabetic? (char -> bool)                                     )
    (char-numeric?  (char -> bool)                                       )
    (char-whitespace? (char -> bool)                                     )
    (char-upper-case? (char -> bool)                                     )
    (char-lower-case? (char -> bool)                                     )
    (char->integer  (char -> num)                                        )
    (integer->char  (num -> char)                                        )
    (char-upcase    (char -> char)                                       )
    (char-downcase  (char -> char)                                       )

    ; strings
    (make-string    (num (&optional char) -> str)                        )
    (string         ((&list char) -> str)                                )
    (string-length  (str -> num)                                         )
    (string-ref     (str num -> char)                                    )
    (string-set!    (str num char -> void)                               )
    (string=?       (str str -> bool)                                    )
    (string<?       (str str -> bool)                                    )
    (string>?       (str str -> bool)                                    )
    (string<=?      (str str -> bool)                                    )
    (string>=?      (str str -> bool)                                    )
    (string-ci=?    (str str -> bool)                                    )
    (string-ci<?    (str str -> bool)                                    )
    (string-ci>?    (str str -> bool)                                    )
    (string-ci<=?   (str str -> bool)                                    )
    (string-ci>=?   (str str -> bool)                                    )
    (substring      (str num num -> str)                                 )
    (string-append  ((&list str) -> str)                                 )
    (string->list   (str -> (list char))                                 (Ic))
    (list->string   ((list char) -> str)                                 )
    (string-copy    (str -> str)                                         )
    (string-fill!   (str char -> void)                                   )

    ; vectors
    (make-vector    (num a -> (vec a))                                   (I)) ; different semantics
    (vector         ((&list a) -> (vec a))                               (I))
    (vector-length  ((vec a) -> num)                                     )
    (vector-ref     ((vec a) num -> a)                                   )
    (vector-set!    ((vec a) num a -> void)                              )
    (vector->list   ((vec a) -> (list a))                                (Ic))
    (list->vector   ((list a) -> (vec a))                                (I))
    (vector-fill!   ((vec a) a -> void)                                  )

    ; control features
    (apply          (((&list a) -> b) (list a) -> b)                     (I) (D))
    (map            ((a -> b) (list a) -> (list b))                      (I) (D))
    (for-each       ((a -> b) (list a) -> void)                          (I) (D))
    (force          ((promise a) -> a)                                   (I))
    (call-with-current-continuation (((a -> b) -> a) -> a)               (I))

    ; input and output
    (call-with-input-file  (str (iport -> a) -> a)                       (I))
    (call-with-output-file (str (oport -> a) -> a)                       (I))
    (input-port?           (a -> bool)                                   )
    (output-port?          (a -> bool)                                   )
    (current-input-port    (-> iport)                                    )
    (current-output-port   (-> oport)                                    )
    (with-input-from-file  (str (-> a) -> a)                             (I))
    (with-output-to-file   (str (-> a) -> a)                             (I))
    (open-input-file       (str -> iport)                                )
    (open-output-file      (str -> oport)                                )
    (close-input-port      (iport -> void)                               )
    (close-output-port     (oport -> void)                               )
    ; this type takes advantage of wrong scope of MU to get more sharing
    (read                  ((&optional iport)
                            ->
                            (+ eof
                               num nil false true char sym str
                               (box (MU sexp
                                        (+ num nil false true char sym str
                                           (vec sexp)
                                           (cons sexp sexp)
                                           (box sexp))))
                               (cons sexp sexp)
                               (vec sexp)))                                (I))
    (read-char             ((&optional iport) -> (+ char eof))             (I))
    (peek-char             ((&optional iport) -> (+ char eof))             (I))
    (char-ready?           ((&optional iport) -> bool)                     (I))
    (write                 (a (&optional oport) -> void)                   (I))
    (display               (a (&optional oport) -> void)                   (I))
    (newline               ((&optional oport) -> void)                     (I))
    (write-char            (char (&optional oport) -> void)                (I))

    ; system interface
    (load           (str -> void)                                          )
    (transcript-on  (str -> void)                                          )
    (transcript-off (-> void)                                              )

    ; non R4RS extensions

    ; misc
    (symbol-append  ((&rest a) -> sym)                                     )
    (box            (a -> (box a))                                         (I))
    (unbox          ((box a) -> a)                                         (S boxX))
    (set-box!       ((box a) a -> void)                                    )
    (void           (-> void)                                              )
    (make-module    (a -> (module a))                                      )
    (raise          ((&rest a) -> b)                                       )

    ; for match
    (match:error    (a (&rest b) -> c)                                     )
    (should-never-reach (a -> b)                                           )

    ; constant vectors
    (make-cvector   (num a -> (cvec a))                                    )
    (cvector        ((&list a) -> (cvec a))                                )
    (cvector-length ((cvec a) -> num)                                      )
    (cvector-ref    ((cvec a) num -> a)                                    )
    (cvector->list  ((cvec a) -> (list a))                                 (Ic))
    (list->cvector  ((list a) -> (cvec a))                                 )

    ; for quoted lists
    (,Qcons         (a b -> (cons a b))                                    (Ic) (N))
    (,Qvector       ((&list a) -> (vec a))                                 (I) (N))
    (,Qbox          (a -> (box a))                                         (I) (N))
    (,Qlist         ((&list a) -> (list a))                                (Ic) (N))

    ; predicates
    (number?        ((+ num x) -> bool)                                  (P (num)))
    (null?          ((+ nil x) -> bool)                                  (P (nil)))
    (char?          ((+ char x) -> bool)                                 (P (char)))
    (symbol?        ((+ sym x) -> bool)                                  (P (sym)))
    (string?        ((+ str x) -> bool)                                  (P (str)))
    (vector?        ((+ (vec a) x) -> bool)                              (P (vec a)))
    (cvector?       ((+ (cvec a) x) -> bool)                             (P (cvec a)))
    (box?           ((+ (box a) x) -> bool)                              (P (box a)))
    (pair?          ((+ (cons a b) x) -> bool)                           (P (cons a b)))
    (procedure?     ((+ ((&rest a) -> b) x) -> bool)                     (P (?-> a b)))
    (eof-object?    ((+ eof x) -> bool)                                  (P (eof)))
    (input-port?    ((+ iport x) -> bool)                                (P (iport)))
    (output-port?   ((+ oport x) -> bool)                                (P (oport)))
    (true-object?   ((+ true x) -> bool)                                 (P (true)))  ; not R4RS
    (false-object?  ((+ false x) -> bool)                                (P (false))) ; not R4RS
    (module?        ((+ (module a) x) -> bool)                           (P (module a))) ; not R4RS
    ;; These last two are not disjoint predicates.
    ;; improve-clauses checks for them explicitly.
    (boolean?       ((+ true false x) -> bool)                           (P #t))
    (list?          ((MU u (+ nil (cons y u) x)) -> bool)                (P #t))
    ))

(define initial-env '())

(define init-env!
  (lambda ()
    (set! initial-env (foldr init-prim empty-env initial-info))))

(define init-prim
  (lambda (l env)
    (letrec ((build-selector
              (match-lambda
                ('X
                  (lambda (X) X))
                ('_
                  (lambda (X) (make-Pany)))
                ('boxX
                  (let ((c (lookup env 'box?)))
                    (lambda (X)
                      (make-Pobj c (list X)))))
                ((x . y)
                 (let ((c (lookup env 'pair?))
                       (lx (build-selector x))
                       (ly (build-selector y)))
                   (lambda (X)
                     (make-Pobj c (list (lx X) (ly X)))))))))
      (match l
        ((name type . attr)
         (let* ((pure (cond ((assq 'I attr) #f)
                            ((assq 'Ic attr) 'cons)
                            (else #t)))
                (def (assq 'D attr))
                (check (assq 'C attr))
                (nocheck (assq 'N attr))
                (pred (match (assq 'P attr)
                        (#f #f)
                        ((_ #t) #t)
                        ((_ (tag . args)) (cons (lookup initial-type-env tag) args))))
                (sel (match (assq 'S attr)
                       (#f #f)
                       ((_ s) (build-selector s))))
                (env1 (extend-env env name
                        (make-Name
                          name
                          (closeall (R+ initial-type-env type))
                          #f
                          0
                          #f
                          #f
                          (cond (nocheck 'nocheck)
                                (check 'check)
                                (def 'imprecise)
                                (else #t))
                          #f
                          pure
                          pred
                          #f
                          sel)))
                (env2 (extend-env env1 (symbol-append 'CHECK- name)
                        (make-Name
                          (symbol-append 'CHECK- name)
                          (closeall (R++ initial-type-env type))
                          #f
                          0
                          #f
                          #f
                          #t
                          #f
                          pure
                          pred
                          #f
                          sel))))
           env2))))))

(define defprim
  (lambda (name type mode)
    (handle
      (R+ initial-type-env type)
      (match-lambda*
        (('type . args) (apply syntax-err type args))
        (x (apply raise x))))
    (let* ((attr (match mode
                   ('impure '((I)))
                   ('pure '())
                   ('pure-if-cons-is '((Ic)))
                   ('mutates-cons
                     (set! cons-mutators (cons name cons-mutators))
                     '())
                   (x (use-error "invalid attribute ~a for st:defprim" x))))
           (info `(,name ,type ,@attr)))
      (unless (equal? info (assq name initial-info))
        (set! initial-info (cons info initial-info))
        (set! initial-env (init-prim info initial-env))))))

(init-types!)
(init-env!)
