;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Soft Scheme - Copyright (c) 1993,1994 by Andrew K. Wright
;; See LICENSE for a full description of the license terms.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Type structures
;;
;; type  ::= (box (C depth:num kind:sym fsym:sym f:flag a:(list type) n:type))
;;         | (box (V depth:num kind:sym name:proc vis:bool split:bool inst:(list type))
;;         | (box 'Bot)
;;         | (box type)
;; flag  ::= (box (V depth:num kind:sym name:proc vis:bool split:bool inst:(list type))
;;         | (box 'Top)
;;         | (box flag)
;;
;; The "depth" field is a number representing the outermost
;; environment variables may appear in; smaller numbers indicate
;; more outer environments, the global environment is ord-depth.
;; Negative numbers indicate generic variables.
;;

(define Populated #t)   ; switch to control treatment of Abs variables
(define Pseudo #f)      ; switch to control pseudo-subtype (default set below)
(define Global-error #f); switch to control global ERROR insertion
(define Share #f)       ; switch to control display of shared flags
(define Matchst #f)     ; switch to control subtyping match input
(define FullSharing #t) ; switch to control how careful generalization is
(define Dump-depths #f) ; switch to control dumping of depths
(define Flags #t)       ; switch to control flagging

(define-structure (C depth kind fsym pres args next))
(define-structure (V depth kind name vis split inst))
(define-structure (TS type n-gen))

(define-structure (K name order args))

(define Top (box 'Top))
(define Bot (box 'Bot))

(define generic?
  (lambda (d)
    (< d 0)))

(define new-type
  (lambda (s d)
    (let ((t (box s)))
      (vector-set! types d (cons t (vector-ref types d)))
      t)))

(define generate-counter
  (lambda ()
    (let ((n 0))
      (lambda ()
        (set! n (+ 1 n))
        n))))

(define var-counter (generate-counter))
(define make-raw-tvar
  (lambda (d k)
    (make-V d k var-counter #t #f #f)))
(define make-tvar
  (lambda (d k)
    (new-type (make-raw-tvar d k) d)))

(define Ord? (lambda (k) (eq? 'Ord k)))
(define Abs? (lambda (k) (eq? 'Abs k)))      ; for procedures
(define Pre? (lambda (k) (eq? 'Pre k)))      ; for records

(define ord-depth 2)

(define depth ord-depth)
(define types (make-vector 16 '()))

(define reset-types!
  (lambda ()
    (set! depth ord-depth)
    (set! types (make-vector 16 '()))))

(define push-level
  (lambda ()
    (set! depth (+ depth 1))
    (when (< (vector-length types) (+ 1 depth))
      (set! types
        (let ((l (vector->list types)))
          (list->vector
            (append l (map (lambda (_) '()) l))))))))
              
(define pop-level
  (lambda ()
    (vector-set! types depth '())
    (set! depth (- depth 1))))

(define V-ord (lambda () (make-tvar depth 'Ord)))
(define V-abs (lambda () (make-tvar depth 'Abs)))
(define V-pre (lambda () (make-tvar depth 'Pre)))

(define Tvar V-ord)
(define out1Tvar (lambda () (make-tvar (- depth 1) 'Ord)))
(define monoTvar (lambda () (make-tvar ord-depth 'Ord)))

(define pvar
  (match-lambda
    (($ box (and x ($ V d k _ vis _ _)))
     (unless (number? (V-name x))
       (set-V-name! x ((V-name x))))
     (string->symbol
       (sprintf "~a~a~a"
         (match k
           ('Ord (if (generic? d)
                     (if vis "X" "x")
                     (if vis "Z" "z")))
           ('Abs (if vis "A" "a"))
           ('Pre (if vis "P" "p")))
         (V-name x)
         (if Dump-depths (sprintf ".~a" d) ""))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Misc useful routines

(define make-tvar-like
  (match-lambda
    (($ box ($ V d k _ _ _ _)) (make-tvar d k))))

(define ind*
  (lambda (t)
    (match (unbox t)
      ((? box? u)
       (let ((v (ind* u)))
         (set-box! t v)
         v))
      (_ t))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Interface routines for dealing with checks

; Is a runtime check caused by this check information?
(define type-check?
  (match-lambda
    ((abs def inexhaust once _)
     (cond (((if once check-abs1? check-abs?) abs)
            (if (and def (definite? def))
                'DEF
                #t))
           (inexhaust 'INEXHAUST)
           (else #f)))))

; Is a runtime check _immediately_ caused by this check information?
(define type-check1?
  (match-lambda
    ((abs def inexhaust _ _)
     (cond ((check-abs1? abs)
            (if (and def (definite? def))
                'DEF
                #t))
           (inexhaust 'INEXHAUST)
           (else #f)))))

; Is a runtime check generated by this variable list?
(define check-abs?
  (lambda (vlist)
    (letrec
      ((seen '())
       (Labs?
        (lambda (t)
          (match t
            (($ box ($ V _ _ _ _ _ inst))
             (and inst (not (memq t seen))
                  (begin
                    (set! seen (cons t seen))
                    (ormap (match-lambda ((t . _) (Labs? t))) inst))))
            (($ box ($ C _ _ _ p _ n))
             (or (Labs? p) (Labs? n)))
            (($ box (? symbol?)) #t)
            (($ box i) (Labs? i))))))
      (ormap Labs? vlist))))

; Is a runtime check _immediately_ generated by this variable list?
(define check-abs1?
  (lambda (vlist)
    (letrec
      ((Labs1?
        (lambda (t)
          (match t
            (($ box (? V?)) #f)
            (($ box ($ C _ _ _ p _ n))
             (or (Labs1? p) (Labs1? n)))
            (($ box (? symbol?)) #t)
            (($ box i) (Labs1? i))))))
      (ormap Labs1? vlist))))

; Return a set of non-local sources of a check.
(define check-sources
  (lambda (info)
    (letrec
      ((seen '())
       (Lsrcs
        (lambda (t source)
          (match t
            (($ box ($ V _ k _ _ _ inst))
             (union
               (if (and inst (not (memq t seen)))
                   (begin
                     (set! seen (cons t seen))
                     (foldr union empty-set
                       (map (match-lambda ((t . s) (Lsrcs t s))) inst)))
                   empty-set)))
            (($ box ($ C _ _ _ p _ n))
             (union (Lsrcs p source) (Lsrcs n source)))
            (($ box (? symbol?)) (if source (set source) empty-set))
            (($ box i) (Lsrcs i source))))))
      (match-let (((abs _ _ _ _) info))
        (if (eq? #t abs)
            empty-set
            (foldr union empty-set
              (map (lambda (t) (Lsrcs t #f)) abs)))))))

; Return a list of local component sources of a check.
(define check-local-sources
  (match-lambda
    ((_ _ _ _ component) component)))

;; Build information about non-absent arguments of a primitive.
;; Returns a value of type:
;;  (MU or (list (+ ty? (cons ty? (cons (list ty?) (cons or nil))))))
(define mk-definite-prim
  (match-lambda
    (($ box ($ C _ _ x p a n))
     (if (eq? (K-name x) '?->)
         (let ((seen '()))
           (recur Lprim ((t (car a)))
             (match t
               (($ box ($ C _ _ x p a n))
                (if (memq t seen)
                    '()
                    (begin
                      (set! seen (cons t seen))
                      (match (K-name x)
                        ('noarg (cons p (Lprim n)))
                        ('arg (let ((args
                                     (recur argloop ((a (car a)))
                                       (match a
                                         (($ box ($ C _ _ _ p _ n))
                                          (cons p (argloop n)))
                                         (($ box ($ V _ k _ _ _ _))
                                          (if (Ord? k) (list a) '()))
                                         (($ box (? symbol?)) '())
                                         (($ box i) (argloop i))))))
                                (cons (list p args (Lprim (cadr a)))
                                      (Lprim n))))))))
               (($ box ($ V _ k _ _ _ _))
                (if (Ord? k)
                    (list t)
                    '()))
               (($ box (? symbol?)) '())
               (($ box i) (Lprim i)))))
         (mk-definite-prim n)))
    (($ box (? V?)) '())
    (($ box (? symbol?)) '())
    (($ box i) (mk-definite-prim i))))

(define mk-definite-app
  (match-lambda
    (($ box ($ C _ _ _ p _ _)) (list p))))

(define mk-definite-lam
  (match-lambda
    (($ box ($ C _ _ x p a n))
     (if (eq? (K-name x) '?->)
         (let ((seen '()))
           (recur Llam ((t (car a)))
             (match t
               (($ box ($ C _ _ x p a n))
                (if (memq t seen)
                    '()
                    (begin
                      (set! seen (cons t seen))
                      (match (K-name x)
                        ('noarg (cons p (Llam n)))
                        ('arg (let ((args (list Top)))
                                (cons (list p args (Llam (cadr a)))
                                      (Llam n))))))))
               (($ box ($ V _ k _ _ _ _))
                (if (Ord? k)
                    (list t)
                    '()))
               (($ box (? symbol?)) '())
               (($ box i) (Llam i)))))
         (mk-definite-lam n)))
    (($ box (? V?)) '())
    (($ box (? symbol?)) '())
    (($ box i) (mk-definite-lam i))))
  
(define definite?
  (lambda (def-info)
    (letrec ((non-empty?
              (lambda (t)
                (let ((seen '()))
                  (recur Ldef ((t t))
                    (match t
                      (($ box ($ C _ _ _ p _ n))
                       (or (Ldef p) (Ldef n)))
                      (($ box ($ V d k _ _ _ inst))
                       (if (or Global-error (Abs? k))
                           (and inst
                                (generic? d)
                                (not (memq t seen))
                                (begin
                                  (set! seen (cons t seen))
                                  (ormap (match-lambda ((t . _) (Ldef t)))
                                    inst)))
                           (generic? d)))
                      (($ box 'Top) #t)
                      (($ box 'Bot) #f)
                      (($ box i) (Ldef i)))))))
             (ok
              (lambda (l)
                (ormap (match-lambda
                         ((? box? t) (non-empty? t))
                         ((p arg rest) (and (non-empty? p)
                                            (ormap non-empty? arg)
                                            (ok rest))))
                  l))))
      (not (ok def-info)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Generalize a type scheme

; Close over type variables of depth at least "depth" in a list of types
(define close 
  (lambda (t-list)
    (close-type t-list #f)))

; Close all variables in a single type
(define closeall
  (lambda (t)
    (car (close-type (list t) #t))))

(define for
  (lambda (from to f)
    (cond ((= from to) (f from))
          ((< from to) (begin (f from) (for (+ from 1) to f)))
          (else        #f))))

(define close-type
  (lambda (t-list all?)
    (let* ((sorted (make-vector (+ depth 1) '()))
           (sort (lambda (t)
                   (match t
                     (($ box ($ C d _ _ _ _ _))
                      (vector-set! sorted d (cons t (vector-ref sorted d))))
                     (($ box ($ V d _ _ _ _ _))
                      (vector-set! sorted d (cons t (vector-ref sorted d))))
                     (_ #f))))
           (prop-d (lambda (down)
                     (letrec ((pr (match-lambda
                                    (($ box (and x ($ V d _ _ _ _ _)))
                                     (when (< down d)
                                       (set-V-depth! x down)))
                                    (($ box (and x ($ C d _ _ p a n)))
                                     (when (< down d)
                                       (set-C-depth! x down)
                                       (pr p)
                                       (for-each pr a)
                                       (pr n)))
                                    (($ box (? symbol?)) #f)
                                    (z (pr (ind* z))))))
                       (match-lambda
                         (($ box (and x ($ C d _ _ p a n)))
                          (when (<= down d)
                            (pr p)
                            (for-each pr a)
                            (pr n)))
                         (_ #f)))))
           (prop-k (lambda (t)
                     (let ((pk (lambda (kind)
                                 (rec pr
                                      (match-lambda
                                        (($ box (and x ($ V _ k _ _ _ _)))
                                         (when (kind< kind k)
                                           (set-V-kind! x kind)))
                                        (($ box (and x ($ C _ k _ p a n)))
                                         (when (kind< kind k)
                                           (set-C-kind! x kind)
                                           (pr p)
                                           (unless Populated
                                             (for-each pr a))
                                           (pr n)))
                                        (($ box (? symbol?)) #f)
                                        (z (pr (ind* z))))))))
                       (match t
                         (($ box (and x ($ C _ k _ p a n)))
                          (when (not (Ord? k))
                            (let ((prop (pk k)))
                              (prop p)
                              (unless Populated
                                (for-each prop a))
                              (prop n))))
                         (_ #f)))))
           (might-be-generalized? (match-lambda
                                    (($ box ($ V d k _ _ _ _))
                                     (and (<= depth d)
                                          (or Populated (Ord? k) all?)))
                                    (($ box ($ C d k _ _ _ _))
                                     (and (<= depth d)
                                          (or Populated (Ord? k) all?)))
                                    (($ box (? symbol?)) #f)))
           (leaves '())
           (depth-of (match-lambda
                       (($ box ($ V d _ _ _ _ _)) d)
                       (($ box ($ C d _ _ _ _ _)) d)))
           (vector-grow (lambda (v)
                          (let* ((n (vector-length v))
                                 (v2 (make-vector (* n 2) '())))
                            (recur loop ((i 0))
                              (when (< i n)
                                (vector-set! v2 i (vector-ref v i))
                                (loop (+ 1 i))))
                            v2)))
           (parents (make-vector 64 '()))
           (parent-index 0)
           (parents-of (lambda (t)
                         (let ((d (depth-of t)))
                           (if (< depth d)
                               (vector-ref parents (- (- d depth) 1))
                               '()))))
           (xtnd-parents! (lambda (t parent)
                            (match t
                              (($ box (and x ($ V d _ _ _ _ _)))
                               (when (= d depth)
                                 (set! parent-index (+ 1 parent-index))
                                 (set-V-depth! x (+ depth parent-index))
                                 (when (< (vector-length parents) parent-index)
                                   (set! parents (vector-grow parents)))
                                 (set! d (+ depth parent-index)))
                               (vector-set! parents (- (- d depth) 1)
                                 (cons parent
                                       (vector-ref parents (- (- d depth) 1)))))
                              (($ box (and x ($ C d _ _ _ _ _)))
                               (when (= d depth)
                                 (set! parent-index (+ 1 parent-index))
                                 (set-C-depth! x (+ depth parent-index))
                                 (when (< (vector-length parents) parent-index)
                                   (set! parents (vector-grow parents)))
                                 (set! d (+ depth parent-index)))
                               (vector-set! parents (- (- d depth) 1)
                                 (cons parent
                                       (vector-ref parents (- (- d depth) 1))))))))
           (needs-cleanup '())
           (revtype (rec revtype
                         (lambda (parent t)
                           (let ((t (ind* t)))
                             (cond ((not (might-be-generalized? t))
                                    #f)
                                   ((null? (parents-of t))
                                    (xtnd-parents! t parent)
                                    (set! needs-cleanup
                                      (cons t needs-cleanup))
                                    (match t
                                      (($ box (? V?))
                                       (set! leaves (cons t leaves)))
                                      (($ box ($ C _ _ _ p a n))
                                       (let ((rev (lambda (q) (revtype t q))))
                                         (rev p)
                                         (for-each rev a)
                                         (rev n)))))
                                   ((not (memq parent (parents-of t)))
                                    (xtnd-parents! t parent))
                                   (else #f))))))
           (generic-index 0)
           (gen (rec gen
                     (lambda (t)
                       (let ((t (ind* t)))
                         (when (might-be-generalized? t)
                           (set! generic-index (- generic-index 1))
                           (let ((parents (parents-of t)))
                             (match t
                               (($ box (and x ($ V _ k _ _ _ _)))
                                (set-V-depth! x generic-index)
                                (when (and Populated
                                           (or Global-error (Abs? k) (Pre? k))
                                           (not all?))
                                  (set-V-inst! x '())))
                               (($ box (? C? x))
                                (set-C-depth! x generic-index)))
                             (for-each gen parents)))))))
           (cleanup (match-lambda
                      (($ box (and x ($ V d _ _ _ _ _)))
                       (unless (< d 0)
                         (set-V-depth! x (- depth 1))))
                      (($ box (and x ($ C d _ _ _ _ _)))
                       (unless (< d 0)
                         (set-C-depth! x (- depth 1))))))
           (gen2 (rec gen    ; old version that does less sharing
                      (lambda (t)
                        (let ((t (ind* t)))
                          (when (might-be-generalized? t)
                            (set! generic-index (- generic-index 1))
                            (match t
                              (($ box (and x ($ V _ k _ _ _ _)))
                               (set-V-depth! x generic-index)
                               (when (and Populated
                                          (or Global-error (Abs? k) (Pre? k))
                                          (not all?))
                                 (set-V-inst! x '())))
                              (($ box (and x ($ C _ _ _ p a n)))
                               (set-C-depth! x generic-index)
                               (gen p)
                               (for-each gen a)
                               (gen n))))))))
           (upd (lambda (t)
                  (let ((d (depth-of t)))
                    (when (< 0 d)
                      (vector-set! types d (cons t (vector-ref types d))))))))
      (for-each sort (vector-ref types depth))
      (for 0 (- depth 1)
           (lambda (i) (for-each (prop-d i) (vector-ref sorted i))))
      (for-each prop-k (vector-ref types depth))
      (vector-set! types depth '())
      (if FullSharing
          (begin
            (for-each (lambda (t) (revtype t t)) t-list)
            (for-each gen leaves)
            (for-each cleanup needs-cleanup))
          (for-each gen2 t-list))
      (for 0 depth
           (lambda (i) (for-each upd (vector-ref sorted i))))
      (if (null? t-list)
          '()
          (match-let* ((n-gen (- generic-index))
                       ((t-list n-gen) (if (and Pseudo Flags (not all?))
                                           (Pseudo t-list n-gen)
                                           (list t-list n-gen))))
            (visible t-list n-gen)
            (map (lambda (t) (make-TS t n-gen)) t-list))))))

(define visible-time 0)

;; Renumber generalized type variables to small numbers and
;; set visible bit for vars that are not useless.
(define visible
  (lambda (t-list n-gen)
    (let* ((before (cpu-time))
           (valences (make-vector n-gen '()))
           (namer (generate-counter))
           (Lvis
            (rec Lvis
                 (lambda (t pos rcd)
                   (match t
                     (($ box ($ C d _ x p a n))
                      (when (and (generic? d)
                                 (not (element-of? pos (vector-ref valences (- (- d) 1)))))
                        (let ((u (union (vector-ref valences (- (- d) 1)) (set pos))))
                          (vector-set! valences (- (- d) 1) u))
                        (Lvis p pos rcd)
                        (match (K-name x)
                          ('?->
                            (Lvis (car a) (not pos) #f)
                            (Lvis (cadr a) pos #f))
                          ('record
                            (Lvis (car a) pos #t))
                          (_
                           (for-each (lambda (x) (Lvis x pos #f)) a)))
                        (Lvis n pos rcd)))
                     (($ box (and x ($ V d k _ _ _ _)))
                      (when (and (generic? d)
                                 (not (element-of? pos (vector-ref valences (- (- d) 1)))))
                        (let ((u (union (vector-ref valences (- (- d) 1)) (set pos))))
                          (vector-set! valences (- (- d) 1) u)
                          (set-V-name! x namer)
                          (cond ((Abs? k)
                                 #f)
                                ((= 2 (cardinality u))
                                 (set-V-split! x #t)
                                 (set-V-vis! x #t))
                                ((eq? pos rcd)
                                 (set-V-vis! x #t))
                                (else
                                 (set-V-vis! x #f))))))
                     (($ box (? symbol?)) #f)
                     (($ box i) (Lvis i pos rcd)))))))
      (for-each (lambda (t) (Lvis t #t #f)) t-list)
      (set! visible-time (+ visible-time (- (cpu-time) before))))))

(define visible?
  (match-lambda
    (($ box ($ V _ k _ vis _ _)) (or (Pre? k) (and vis (not (Abs? k)))))
    (($ box 'Top) #t)
    (($ box 'Bot) #f)
    (($ box i) (visible? i))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Instantiate a type scheme, replacing bound type variables by fresh ones.
;;
;; Returns a list consisting of the instantiated type and a list
;; of the new Abs and Pre variables

(define instantiate
  (lambda (ts syntax)
    (match ts
      (($ TS t n-gen)
       (let*
         ((absv '())
          (seen (make-vector n-gen #f))
          (t2 (recur Linst ((t t))
                (match t
                  (($ box (and y ($ V d k _ _ _ inst)))
                   (cond ((not (generic? d)) t)
                         ((vector-ref seen (- (- d) 1)))
                         (else
                          (let ((u (make-tvar depth k)))
                            (vector-set! seen (- (- d) 1) u)
                            (when inst
                              (set-V-inst! y (cons (cons u syntax) inst)))
                            (when (or (Abs? k) (Pre? k))
                              (set! absv (cons u absv)))
                            u))))
                  (($ box ($ C d _ x p a n))
                   (cond ((not (generic? d)) t)
                         ((vector-ref seen (- (- d) 1)))
                         (else
                          (let ((u (new-type '**FIX** depth)))
                            (vector-set! seen (- (- d) 1) u)
                            (set-box! u
                              (make-C depth 'Ord x
                                (if Flags (Linst p) Top)
                                (map Linst a)
                                (Linst n)))
                            u))))
                  (($ box (? symbol?)) t)
                  (($ box i) (Linst i))))))
         (list t2 absv))))))


;; FAKE some subtyping at polymorphic generalization by replacing Tops
;; in negative positions with fresh variables, and absent variables in
;; positive positions with fresh variables.
;;
;; This algorithm only replaces things under generic nodes, thus never
;; breaks sharing.  It operates after generalization.
(define pseudo-subtype
  (lambda (t-list n-gen)
     (let* ((valences (make-vector n-gen '()))
            (valence-of
             (lambda (d)
               (vector-ref valences (- (- d) 1))))
            (set-valence
             (lambda (d v)
               (vector-set! valences (- (- d) 1) v)))
            (find
             (rec find
                  (lambda (t pos mutable)
                    (match t
                      (($ box ($ V d _ _ _ _ _))
                       (when (generic? d)
                         (cond
                           (mutable
                            (set-valence d (set #t #f)))
                           ((not (element-of? pos (valence-of d)))
                            (set-valence d (union (valence-of d) (set pos))))
                           (else #f))))
                      (($ box ($ C d _ x p a n))
                       (when (generic? d)
                         (cond
                           ((= 2 (cardinality (valence-of d)))
                            #f)
                           (mutable
                            (set-valence d (set #t #f))
                            (for-each2
                              (lambda (t m)
                                (find t pos mutable))
                              a (K-args x))
                            (find n pos mutable))
                           ((not (element-of? pos (valence-of d)))
                            (set-valence d (union (valence-of d) (set pos)))
                            (if (eq? '?-> (K-name x))
                                (begin
                                  (find (car a) (not pos) mutable)
                                  (find (cadr a) pos mutable))
                                (for-each2
                                  (lambda (t m)
                                    (find t pos (or m mutable)))
                                  a (K-args x)))
                            (find n pos mutable))
                           (else #f))))
                    (($ box (? symbol?)) #f)
                    (($ box i) (find i pos mutable))))))
            (seen (make-vector n-gen #f))
            (new-generic-var
             (lambda ()
               (set! n-gen (+ 1 n-gen))
               (box (make-raw-tvar (- n-gen) 'Ord))))
            (copy
             (rec copy
                  (lambda (t)
                    (match t
                      (($ box ($ V d k _ _ _ _))
                       (if (generic? d)
                           (or (vector-ref seen (- (- d) 1))
                               (let ((u (if (and (Abs? k)
                                                 (equal? (valence-of d) '(#t)))
                                            (new-generic-var)
                                            t)))
                                 (vector-set! seen (- (- d) 1) u)
                                 u))
                           t))
                      (($ box ($ C d k x p a n))
                       (if (generic? d)
                           (or (vector-ref seen (- (- d) 1))
                               (let* ((u (box '**FIX**))
                                      (_ (vector-set! seen (- (- d) 1) u))
                                      (new-p (if (and (eq? (ind* p) Top)
                                                      (equal? (valence-of d) '(#f)))
                                                 (new-generic-var)
                                                 (copy p)))
                                      (new-a (map copy a))
                                      (new-n (copy n)))
                                 (set-box! u (make-C d 'Ord x new-p new-a new-n))
                                 u))
                           t))
                      (($ box (? symbol?)) t)
                      (($ box i) (copy i))))))
            (t-list (map (lambda (t) (find t #t #f) (copy t)) t-list)))
       (list t-list n-gen))))

(set! Pseudo pseudo-subtype)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Unification

(define unify
  (letrec
    ((uni
      (lambda (u v)
        (unless (eq? u v)
          (match (cons u v)
            ((($ box (and us ($ C ud uk ux up ua un))) .
              ($ box (and vs ($ C vd vk vx vp va vn))))
             (if (eq? ux vx)
                 (begin
                   (if (< ud vd)
                       (begin
                         (set-box! v u)
                         (when (kind< vk uk) (set-C-kind! us vk)))
                       (begin
                         (set-box! u v)
                         (when (kind< uk vk) (set-C-kind! vs uk))))
                   (uni un vn)
                   (for-each2 uni ua va)
                   (uni up vp))
                 (let* ((next (Tvar))
                        (k (if (kind< uk vk) uk vk)))
                   (if (< ud vd)        ; (K< ux vx)
                       (begin
                         (when (< vd ud) (set-C-depth! us vd))
                         (when (kind< vk uk) (set-C-kind! us vk))
                         (set-box! v u))
                       (begin
                         (when (< ud vd) (set-C-depth! vs ud))
                         (when (kind< uk vk) (set-C-kind! vs uk))
                         (set-box! u v)))
                   (uni (new-type (make-C depth k ux up ua next) depth) vn)
                   (uni un (new-type (make-C depth k vx vp va next) depth)))))
            ((($ box (and x ($ V ud uk _ _ _ _))) .
              ($ box        ($ V vd vk _ _ _ _)))
             (set-V-depth! x (min ud vd))
             (set-V-kind! x (if (kind< uk vk) uk vk))
             (set-box! v u))
            ((($ box ($ V ud uk _ _ _ _)) . ($ box (and x ($ C vd vk _ _ _ _))))
             (when (< ud vd) (set-C-depth! x ud))
             (when (kind< uk vk) (set-C-kind! x uk))
             (set-box! u v))
            ((($ box (and x ($ C ud uk _ _ _ _))) . ($ box ($ V vd vk _ _ _ _)))
             (when (< vd ud) (set-C-depth! x vd))
             (when (kind< vk uk) (set-C-kind! x vk))
             (set-box! v u))
            ((($ box ($ V _ _ _ _ _ _)) . ($ box (? symbol?)))
             (set-box! u v))
            ((($ box (? symbol?)) . ($ box ($ V _ _ _ _ _ _)))
             (set-box! v u))
            ((($ box 'Bot) . ($ box ($ C _ _ _ p _ n)))
             (set-box! v u)
             (uni u p)
             (uni u n))
            ((($ box ($ C _ _ _ p _ n)) . ($ box 'Bot))
             (set-box! u v)
             (uni v p)
             (uni v n))
            (_
             (uni (ind* u) (ind* v))))))))
    uni))

(define kind<
  (lambda (k1 k2)  ; lattice:  Abs < Ord ;  Pre < Ord
    (and (Ord? k2) (not (Ord? k1)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Tidy type encoding

;; Grammar:
;;  type  ::= union | tyvar | (MU tyvar type) | bool | (list type) | k
;;  union ::= (+ k ... tyvar) | (+ k ...)

;; Known problems:
;;  1. MU with recursion thru function arg is wrong (pos).
;;  2. MU name is not unbound outside of definition.
;;  3. let pos is wrong, even without recursion.

(define R+-
  (lambda (flag+ flag- tail+- absent- pos env type)
    (letrec
      ((absent+ V-ord)
       (tvars '())
       (fvars '())
       (absv '())
       (make-flag
        (lambda (pos)
          (cond ((not Flags) Top)
                (pos (flag+))
                (else (flag-)))))
       (typevar?
        (lambda (v)
          (and (symbol? v)
               (not (bound? env v))
               (not (memq v '(_ bool MU list &list &optional &rest arglist + not rec *tidy))))))
       (parse-type
        (lambda (t pos)
          (match t
            (('MU a t)
             (unless (typevar? a)
               (raise 'type "invalid type syntax at ~a" t))
             (when (assq a tvars)
               (raise 'type "~a is defined more than once" a))
             (let* ((fix (new-type '**FIX** depth))
                    (_ (set! tvars (cons (list a fix '()) tvars)))
                    (t (parse-type t pos)))
               (when (eq? t fix)
                 (raise 'type "recursive type is not contractive"))
               (set-box! fix t)
               (ind* t)))
            (('rec (? list? bind) t2)
             (for-each
               (match-lambda
                 ((a _)
                  (unless (typevar? a)
                    (raise 'type "invalid type syntax at ~a" t))
                  (when (assq a tvars)
                    (raise 'type "~a is defined more than once" a))
                  (set! tvars (cons (list a (new-type '**FIX** depth) '()) tvars)))
                 (_
                  (raise 'type "invalid type syntax at ~a" t)))
               bind)
             (for-each
               (match-lambda
                 ((a t)
                  (match (assq a tvars)
                    ((_ fix _)
                     (let ((t (parse-type t '?)))
                       (when (eq? t fix)
                         (raise 'type "type is not contractive"))
                       (set-box! fix t))))))
               bind)
             (parse-type t2 pos))
            ('bool
              (parse-type '(+ false true) pos))
            ('s-exp
              (let ((v (gensym)))
                (parse-type `(MU ,v (+ num
                                       nil
                                       false
                                       true
                                       char
                                       sym
                                       str
                                       (vec ,v)
                                       (box ,v)
                                       (cons ,v ,v)))
                  pos)))
            (('list t)
             (let ((u (gensym)))
               (parse-type `(MU ,u (+ nil (cons ,t ,u))) pos)))
            (('arglist t)
             (let ((u (gensym)))
               (parse-type `(MU ,u (+ noarg (arg ,t ,u))) pos)))
            (('+ . (? list? union))
             (parse-union union pos))
            (t
             (parse-union (list t) pos)))))
       (parse-union
        (lambda (t pos)
          (letrec ((sort-cs
                    (lambda (cs)
                      (sort-list cs (lambda (x y) (K< (C-fsym x) (C-fsym y))))))
                   (link
                    (lambda (c t)
                      (set-C-next! c t)
                      (new-type c depth))))
            (recur loop ((t t)(cs '()))
              (match t
                (()
                 (foldr link
                   (if pos
                       (absent+)
                       (let ((v (absent-)))
                         (set! absv (cons v absv))
                         v))
                   (sort-cs cs)))
                (((? box? t))
                 (foldr link
                   t
                   (sort-cs cs)))
                (('_)
                 (foldr link
                   (tail+-)
                   (sort-cs cs)))
                (((? symbol? a))
                 (=> fail)
                 (unless (typevar? a) (fail))
                 (let* ((cs (sort-cs cs))
                        (ks (map C-fsym cs)))
                   (foldr link
                     (match (assq a tvars)
                       ((_ f aks)
                        (unless (equal? ks aks)
                          (raise 'type "variable ~a is not tidy" a))
                        f)
                       (#f (let ((v (tail+-)))
                             (set! tvars (cons (list a v ks) tvars))
                             v)))
                     cs)))
                ((k . rest)
                 (loop rest (cons (parse-k k pos) cs))))))))
       (parse-k
        (lambda (k pos)
          (cond
            ((and (list? k)
                  (let ((n (length k)))
                    (and (<= 2 n) (eq? '-> (list-ref k (- n 2))))))
             (let* ((rk (reverse k))
                    (arg (reverse (cddr rk)))
                    (res (car rk)))
               (letrec ((mkargs (match-lambda
                                  (() 'noarg)
                                  ((('&rest x)) x)
                                  ((('&list x)) (let ((u (gensym)))
                                                  `(MU ,u (+ noarg (arg ,x ,u)))))
                                  ((('&optional x)) `(+ noarg (arg ,x noarg)))
                                  ((x . y) `(arg ,x ,(mkargs y)))
                                  (_ (raise 'type "invalid type syntax")))))
                 (make-C depth 'Ord
                   (lookup env '?->)
                   (make-flag pos)
                   (let ((a (parse-type (mkargs arg) (flip pos)))
                         (r (parse-type res pos)))
                     (list a r))
                   '**FIX**))))
            (else
             (match k
               ((arg '?-> res)
                (make-C depth 'Ord
                  (lookup env '?->)
                  (make-flag pos)
                  (let ((a (parse-type arg (flip pos)))
                        (r (parse-type res pos)))
                    (list a r))
                  '**FIX**))
               (('record . (? list? fields))
                (make-C depth 'Ord
                  (lookup env 'record)
                  (make-flag pos)
                  (list
                    (recur loop ((fields fields))
                      (match fields
                        (()
                         (if pos Bot (V-ord)))
                        ((((? symbol? f) ftype) . rest)
                         (new-type
                           (make-C depth 'Ord
                             (new-field! f)
                             (if pos
                                 (V-ord)
                                 (let ((v (V-pre)))
                                   (set! absv (cons v absv))
                                   v))
                             (list (parse-type ftype pos))
                             (loop rest))
                           depth)))))
                  '**FIX**))
               (('not (? K? k))
                (make-C depth 'Ord
                  k
                  (if pos
                      (absent+)
                      (let ((v (absent-)))
                        (set! absv (cons v absv))
                        v))
                  (map (lambda (x) (tail+-))
                       (K-args k))
                  '**FIX**))
               (('not c)
                (unless (bound? env c)
                  (raise 'type "invalid type syntax at ~a" k))
                (let ((k (lookup env c)))
                  (make-C depth 'Ord
                    k
                    (if pos
                        (absent+)
                        (let ((v (absent-)))
                          (set! absv (cons v absv))
                          v))
                    (map (lambda (x) (tail+-))
                         (K-args k))
                    '**FIX**)))
               (('*tidy c (? symbol? f))
                (unless (bound? env c)
                  (raise 'type "invalid type syntax at ~a" k))
                (let ((k (lookup env c)))
                  (make-C depth 'Ord
                    k
                    (match (assq f fvars)
                      ((_ . f) f)
                      (#f (let ((v (tail+-))) ; even tho its a flag
                            (set! fvars (cons (cons f v) fvars))
                            v)))
                    (map (lambda (x) (parse-type '(+) pos))
                         (K-args k))
                    '**FIX**)))
               (((? K? k) . (? list? arg))
                (unless (= (length arg) (length (K-args k)))
                  (raise 'type "~a requires ~a arguments"
                    (K-name k) (length (K-args k))))
                (make-C depth 'Ord
                  k
                  (make-flag pos)
                  (smap (lambda (x) (parse-type x pos)) arg)
                  '**FIX**))
               ((c . (? list? arg))
                (unless (bound? env c)
                  (raise 'type "invalid type syntax at ~a" k))
                (let ((k (lookup env c)))
                  (unless (= (length arg) (length (K-args k)))
                    (raise 'type "~a requires ~a arguments"
                      c (length (K-args k))))
                  (make-C depth 'Ord
                    k
                    (make-flag pos)
                    (smap (lambda (x) (parse-type x pos)) arg)
                    '**FIX**)))
               (c
                (unless (bound? env c)
                  (raise 'type "invalid type syntax at ~a" k))
                (let ((k (lookup env c)))
                  (unless (= 0 (length (K-args k)))
                    (raise 'type "~a requires ~a arguments"
                      c (length (K-args k))))
                  (make-C depth 'Ord
                    k
                    (make-flag pos)
                    '()
                    '**FIX**))))))))
       (flip (match-lambda
               ('? '?)
               (#t #f)
               (#f #t))))
      (let ((t (parse-type type pos)))
        (list t absv)))))

(define V-top (lambda () Top))

(define R+   (lambda (env t) (car (R+- V-top V-ord V-ord V-abs #t env t))))
(define R-   (lambda (env t) (car (R+- V-top V-ord V-ord V-abs #f env t))))
(define R++  (lambda (env t) (car (R+- V-top V-ord V-ord V-ord #t env t))))

(define R+collect (lambda (env t) (R+- V-top V-ord V-ord V-abs #t env t)))
(define R-collect (lambda (env t) (R+- V-top V-ord V-ord V-abs #f env t)))

(define R (lambda (t) (R+ initial-type-env t)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Match typing

(define R-match
  (lambda (t)
    (close '())
    '(pretty-print `(FIXING ,(ptype t)))
    (fix-pat-abs! t)
    (list t (collect-abs t))))

(define collect-abs
  (lambda (t)
    (let ((seen '()))
      (recur loop ((t t))
        (match t
          (($ box ($ V _ k _ _ _ _))
           (if (Abs? k) (set t) empty-set))
          (($ box ($ C _ _ _ p a n))
           (if (memq t seen)
               empty-set
               (begin
                 (set! seen (cons t seen))
                 (foldr union (union (loop p) (loop n)) (map loop a)))))
          (($ box (? symbol?)) empty-set)
          (($ box i) (loop i)))))))

; Set variables at current depth to Abs.
; Replace Top flags by variables if Matchst is true.
(define fix-pat-abs!
  (lambda (t)
    (let ((seen '()))
      (recur loop ((t t))
        (match t
          (($ box (and x ($ V d _ _ _ _ _)))
           (when (= d depth)
             (set-V-kind! x 'Abs)))
          (($ box (and c ($ C _ _ _ p a n)))
           (unless (memq t seen)
             (set! seen (cons t seen))
             (loop p)
             (when (and Matchst Flags (eq? (ind* p) Top))
               (set-C-pres! c (V-ord)))
             (for-each loop a)
             (loop n)))
          (($ box (? symbol?)) t)
          (($ box i) (loop i)))))))

; Bind a pattern variable by copying a type, sharing < depth variables.
(define pat-var-bind
  (lambda (t)
    (let ((seen '()))
      (recur loop ((t t))
        (match t
          (($ box ($ V d _ _ _ _ _))
           (if (< d depth)
               t
               (match (assq t seen)
                 ((_ . new) new)
                 (#f (let* ((new (V-ord)))
                       (set! seen (cons (cons t new) seen))
                       new)))))
          (($ box ($ C d k x p a n))
           (match (assq t seen)
             ((_ . new) new)
             (#f (let* ((fix (new-type '**FIX** depth))
                        (fixbox (box fix))
                        (_ (set! seen (cons (cons t fixbox) seen)))
                        (new-p (if Flags (loop p) Top))
                        (new-a (map2 (lambda (mutable a)
                                       (if mutable a (loop a)))
                                     (K-args x)
                                     a))
                        (new-n (loop n)))
                   (if (and ;;  #f ; cocks up typing somehow??
                            (eq? new-p p)
                            (eq? new-n n)
                            (andmap eq? new-a a))
                       (begin
                         (set-box! fixbox t)
                         t)
                       (begin
                         (set-box! fix (make-C d k x new-p new-a new-n))
                         fix))))))
          (($ box (? symbol?)) t)
          (($ box i) (loop i)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Ground types and type constructors
                    
(define fields '())

(define new-field!
  (lambda (x)
    (match (assq x fields)
      (#f (let ((k (make-K
                     x
                     (+ 1 (length fields))
                     '(#f))))
            (set! fields (cons (cons x k) fields))
            k))
      ((_ . k) k))))

(define K<
  (lambda (x y)
    (< (K-order x) (K-order y))))

(define k-counter 0)

(define bind-tycon
  (lambda (x args covers fail-thunk)
    (when (memq x '(_ bool MU list &list &optional &rest arglist + not rec *tidy))
      (fail-thunk "invalid type constructor ~a" x))
    (set! k-counter (+ 1 k-counter))
    (make-K
      (if covers
          (symbol-append x "." (- k-counter 100))
          x)
      k-counter
      args)))

(define initial-type-env '())

(define init-types!
  (lambda ()
    (set! k-counter 0)
    (set! var-counter (generate-counter))
    (set! initial-type-env
      (foldl
        (lambda (l env)
          (extend-env env (car l)
            (bind-tycon (car l) (cdr l) #f (lambda x (apply disaster 'init x)))))
        empty-env
        initial-type-info))
    (set! k-counter 100)
    (reset-types!)))

(define reinit-types!
  (lambda ()
    (set! var-counter (generate-counter))
    (set! k-counter 100)
    (set! fields '())
    (set-cons-mutability! #t)
    (reset-types!)))

(define deftype
  (lambda (tag mutability)
    (set! initial-type-env
      (extend-env initial-type-env tag
        (make-K tag (+ 1 (length initial-type-env)) mutability)))))

(define initial-type-info
  '((?-> #f #f)
    (arg #f #f)
    (noarg)
    (num)
    (nil)
    (false)
    (true)
    (char)
    (sym)
    (str)
    (void)
    (iport)
    (oport)
    (eof)
    (vec #t)
    (box #t)
    (cons #t #t)
    (cvec #f)
    (promise #t)
    (record #f)
    (module #f)))

(define cons-is-mutable #f)

(define set-cons-mutability!
  (lambda (m)
    (set! cons-is-mutable m)
    (set-K-args! (lookup initial-type-env 'cons) (list m m))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; for debugging

(define tidy?
  (lambda (t)
    (let ((seen '()))
      (recur loop ((t t)(label '()))
        (match t
          (($ box (? V?))
           (match (assq t seen)
             (#f (set! seen (cons (cons t label) seen))
                 #t)
             ((_ . l2) (equal? label l2))))
          (($ box ($ C _ _ x _ a n))
           (match (assq t seen)
             ((_ . l2) (equal? label l2))
             (#f (set! seen (cons (cons t label) seen))
                 (and (loop n (sort-list (cons x label) K<))
                      (andmap (lambda (t) (loop t '())) a)))))
          (($ box (? symbol?)) #t)
          (($ box i) (loop i label)))))))
